#!/usr/bin/env python
'''
Modified parameters file for the Hybrid LFP scheme, applying the methodology with
the model of:

Potjans, T. and Diesmann, M. "The Cell-Type Specific Cortical Microcircuit:
Relating Structure and Activity in a Full-Scale Spiking Network Model".
Cereb. Cortex (2014) 24 (3): 785-806.
doi: 10.1093/cercor/bhs358


'''
import numpy as np
import os
import json
from mpi4py import MPI #this is needed to initialize other classes correctly


###################################
# Initialization of MPI stuff     #
###################################
COMM = MPI.COMM_WORLD
SIZE = COMM.Get_size()
RANK = COMM.Get_rank()


####################################
# HELPER FUNCTIONS                 #
####################################

flattenlist = lambda lst: sum(sum(lst, []),[])


####################################
# SPATIAL CONNECTIVITY EXTRACTION  #
####################################

'''
Include functions that extract information from binzegger.json here
'''



def get_F_y(fname='binzegger_connectivity_table.json', y=['p23']): 
    '''
    Extract frequency of occurrences of those cell types that are modeled.
    The data set contains cell types that are not modeled (TCs etc.)
    The returned percentages are renormalized onto modeled cell-types, i.e. they sum up to 1 
    '''
    # Load data from json dictionary
    f = open(fname,'r')
    data = json.load(f)
    f.close()
    
    occurr = []
    for cell_type in y:
        occurr += [data['data'][cell_type]['occurrence']]
    return list(np.array(occurr)/np.sum(occurr)) 



def get_L_yXL(fname, y, x_in_X, L):
    '''
    compute the layer specificity, defined as:
    ::
    
        L_yXL = k_yXL / k_yX
    '''
    def _get_L_yXL_per_yXL(fname, x_in_X, X_index,
                                  y, layer):
        # Load data from json dictionary
        f = open(fname, 'r')
        data = json.load(f)
        f.close()
    
        
        # Get number of synapses
        if layer in [str(key) for key in data['data'][y]['syn_dict'].keys()]:
            #init variables
            k_yXL = 0
            k_yX = 0
            
            for x in x_in_X[X_index]:
                p_yxL = data['data'][y]['syn_dict'][layer][x] / 100.
                k_yL = data['data'][y]['syn_dict'][layer]['number of synapses per neuron']
                k_yXL += p_yxL * k_yL
                
            for l in [str(key) for key in data['data'][y]['syn_dict'].keys()]:
                for x in x_in_X[X_index]:
                    p_yxL = data['data'][y]['syn_dict'][l][x] / 100.
                    k_yL = data['data'][y]['syn_dict'][l]['number of synapses per neuron']
                    k_yX +=  p_yxL * k_yL
            
            if k_yXL != 0.:
                return k_yXL / k_yX
            else:
                return 0.
        else:
            return 0.


    #init dict
    L_yXL = {}

    #iterate over postsynaptic cell types
    for y_value in y:
        #container
        data = np.zeros((len(L), len(x_in_X)))
        #iterate over lamina
        for i, Li in enumerate(L):
            #iterate over presynapse population inds
            for j in range(len(x_in_X)):
                data[i][j]= _get_L_yXL_per_yXL(fname, x_in_X,
                                                          X_index=j,
                                                          y=y_value,
                                                          layer=Li)
        L_yXL[y_value] = data

    return L_yXL



def get_T_yX(fname, y, y_in_Y, x_in_X, F_y):
    '''
    compute the cell type specificity, defined as:
    ::
    
        T_yX = K_yX / K_YX
            = F_y * k_yX / sum_y(F_y*k_yX) 
    
    '''
    def _get_k_yX_mul_F_y(y, y_index, X_index):
        # Load data from json dictionary
        f = open(fname, 'r')
        data = json.load(f)
        f.close()
    
        #init variables
        k_yX = 0.
        
        for l in [str(key) for key in data['data'][y]['syn_dict'].keys()]:
            for x in x_in_X[X_index]:
                p_yxL = data['data'][y]['syn_dict'][l][x] / 100.
                k_yL = data['data'][y]['syn_dict'][l]['number of synapses per neuron']
                k_yX +=  p_yxL * k_yL
        
        return k_yX * F_y[y_index]


    #container
    T_yX = np.zeros((len(y), len(x_in_X)))
    
    #iterate over postsynaptic cell types
    for i, y_value in enumerate(y):
        #iterate over presynapse population inds
        for j in range(len(x_in_X)):
            k_yX_mul_F_y = 0
            for k, yy in enumerate(sum(y_in_Y, [])):                
                if y_value in yy:
                    for yy_value in yy:
                        ii = np.where(np.array(y) == yy_value)[0][0]
                        k_yX_mul_F_y += _get_k_yX_mul_F_y(yy_value, ii, j)
            
            
            if k_yX_mul_F_y != 0:
                T_yX[i, j] = _get_k_yX_mul_F_y(y_value, i, j) / k_yX_mul_F_y
            
    return T_yX


class general_params(object):
    '''class defining general model parameters'''
    def __init__(self):
        '''class defining general model parameters'''

        ####################################
        # REASON FOR THIS SIMULATION       #
        ####################################

        self.reason = 'Default Potjans model with spontaneous activity'

    ####################################
    #                                  #
    #                                  #
    #     SIMULATION PARAMETERS        #
    #                                  #
    #                                  #
    ####################################  


        ####################################
        # MAIN SIMULATION CONTROL          #
        ####################################

        # simulation step size
        self.dt = 0.1

        # simulation start
        self.tstart = 0

        # simulation stop
        self.tstop = 1200


        ####################################
        # OUTPUT LOCATIONS                 #
        ####################################
        
        # TODO: try except does not work with hambach

        # folder for all simulation output and scripts
        # here, compute clusters have scratch areas for saving
        if os.path.isdir(os.path.join('/', 'scratch', os.environ['USER'])):
            self.savefolder = os.path.join('/', 'scratch', os.environ['USER'],
                                           'hybrid_model',
                                           'simulation_output_example_microcircuit')
        # LOCALLY
        else:
            self.savefolder = 'simulation_output_example_microcircuit'

        # folder for simulation scripts
        self.sim_scripts_path = os.path.join(self.savefolder, 'sim_scripts')

        # folder for each individual cell's output
        self.cells_path = os.path.join(self.savefolder, 'cells')

        # folder for figures
        self.figures_path = os.path.join(self.savefolder, 'figures')

        # folder for population resolved output signals
        self.populations_path = os.path.join(self.savefolder, 'populations')

        # folder for raw nest output files
        self.raw_nest_output_path = os.path.join(self.savefolder,
                                                 'raw_nest_output')
        
        # folder for processed nest output files
        self.spike_output_path = os.path.join(self.savefolder,
                                              'processed_nest_output')

    
    ####################################
    #                                  #
    #                                  #
    #        MODEL PARAMETERS          #
    #                                  #
    #                                  #
    ####################################



        ####################################
        # POPULATIONS                      #
        ####################################

        # Number of populations
        self.Npops = 9

        # number of neurons in each population (unscaled)
        self.full_scale_num_neurons = [[20683,   # layer 23 e
                                        5834 ],  # layer 23 i
                                       [21915,   # layer 4 e
                                        5479 ],  # layer 4 i
                                       [4850,    # layer 5 e
                                        1065 ],  # layer 5 i
                                       [14395,   # layer 6 e
                                        2948 ]]  # layer 6 i

        # Number of thalamic neurons/ point processes
        self.n_thal = 902

        # population names TODO: rename
        self.X = ['TC', 'L23E', 'L23I', 'L4E', 'L4I', 'L5E', 'L5I', 'L6E', 'L6I']
        self.Y = self.X[1:]

        # TC and cortical population sizes in one list TODO: rename
        self.N_X = np.array([self.n_thal]+flattenlist([self.full_scale_num_neurons]))
  

        ####################################
        # CONNECTIVITY                     #
        ####################################
        
        # intra-cortical connection probabilities between populations
        #                            23e      23i      4e     4i      5e     5i       6e      6i
        self.conn_probs = np.array([[0.1009,0.1689, 0.0437, 0.0818, 0.0323, 0.,     0.0076,  0.    ],  # 23e
                                    [0.1346,0.1371, 0.0316, 0.0515, 0.0755, 0.,     0.0042,  0.    ],  # 23i
                                    [0.0077,0.0059, 0.0497, 0.135,  0.0067, 0.0003, 0.0453,  0.    ],  # 4e
                                    [0.0691,0.0029, 0.0794, 0.1597, 0.0033, 0.,     0.1057,  0.    ],  # 4i
                                    [0.1004,0.0622, 0.0505, 0.0057, 0.0831, 0.3726, 0.0204,  0.    ],  # 5e 
                                    [0.0548,0.0269, 0.0257, 0.0022, 0.06,   0.3158, 0.0086,  0.    ],  # 5i
                                    [0.0156,0.0066, 0.0211, 0.0166, 0.0572, 0.0197, 0.0396,  0.2252],  # 6e
                                    [0.0364,0.001,  0.0034, 0.0005, 0.0277, 0.008,  0.0658,  0.1443]]) # 6i

        # connection probabilities for thalamic input
        self.C_th = [[0.0,       # layer 23 e
                      0.0   ],   # layer 23 i    
                     [0.0983,    # layer 4 e
                      0.0619],   # layer 4 i
                     [0.0,       # layer 5 e
                      0.0   ],   # layer 5 i
                     [0.0512,    # layer 6 e
                      0.0196]]   # layer 6 i
             


        # full connection probabilities including TC connections
        self.C_YX = np.c_[flattenlist([self.C_th]), self.conn_probs]
        

        ####################################
        # CONNECTION PROPERTIES            #
        ####################################
                
        # mean EPSP amplitude (mV) for all connections except L4e->L23e
        self.PSP_e = 0.15

        # mean EPSP amplitude (mv) for L4e->L23e connections
        # FIX POLISH NOTATION !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        self.PSP_23e_4e = self.PSP_e*2

        # standard deviation of PSC amplitudes relative to mean PSC amplitudes
        # this is sigma/mu in probability distribution
        # Gaussian (lognormal_weights = False): mu is mean, sigma is standard deviation
        # Lognormal (lognormal_weights = False): mean and stdev can be calculated from mu and sigma
        self.PSC_rel_sd = 0.1
                
        # IPSP amplitude relative to EPSP amplitude
        self.g = -4.               

        # set L4i ->L4e stronger in order to get rid of 84 Hz peak
        self.g_4e_4i = self.g*1.15
        
        # Whether to use lognormal weights or not
        self.lognormal_weights = False 

        # mean dendritic delays for excitatory and inhibitory transmission (ms)
        self.delays = [1.5, 0.75] 

        # standard deviation relative to mean delays
        self.delay_rel_sd = 0.5
        
      
        ####################################
        # CELL-TYPE PARAMETERS             #
        ####################################
        
        # Note that these parameters are only relevant for the point-neuron network in case
        # one wants to calculate depth-resolved cell-type specific input currents

        # point to .json connectivity table file
        self.connectivity_table = 'binzegger_connectivity_table.json'

        #list of cell type names used in this script
        #names of every post-syn pop layer
        self.y_in_Y = [
                [['p23'],['b23','nb23']],
                [['p4','ss4(L23)','ss4(L4)'],['b4','nb4']],
                [['p5(L23)','p5(L56)'],['b5','nb5']],
                [['p6(L4)','p6(L56)'],['b6','nb6']]]

        self.y = flattenlist(self.y_in_Y)
        
        #need presynaptic cell type to population mapping
        self.x_in_X = [['TCs', 'TCn']] + sum(self.y_in_Y, [])
        
        
        #map the pre-synaptic populations to the post-syn populations
        self.mapping_Yy = zip(
                  ['L23E', 'L23I', 'L23I',
                   'L4E', 'L4E', 'L4E', 'L4I', 'L4I',
                   'L5E', 'L5E', 'L5I', 'L5I',
                   'L6E', 'L6E', 'L6I', 'L6I'],
                  self.y)

        # Frequency of occurrence of each cell type (F_y); 1-d array 
        self.F_y = get_F_y(fname=self.connectivity_table, y=self.y)

        # Relative frequency of occurrence of each cell type within its population (F_{y,Y})
        self.F_yY = [[get_F_y(fname=self.connectivity_table, y=y) for y in Y] for Y in self.y_in_Y]
        
        # Number of neurons of each cell type (N_y); 1-d array
        self.N_y =  np.array([self.full_scale_num_neurons[layer][pop] * self.F_yY[layer][pop][k] \
                                     for layer, array in enumerate(self.y_in_Y)\
                                     for pop, cell_types in enumerate(array) \
                                     for k, _ in enumerate(cell_types)]).astype(int)
        
        
        #compute the number of synapses as in Potjans&Diesmann 2012
        K_YX = np.zeros(self.C_YX.shape)
        for i in xrange(K_YX.shape[1]):
            K_YX[:, i] = (np.log(1. - self.C_YX[:, i]) /
                                np.log(1. - 1./(self.N_X[1:]*
                                                self.N_X[i])))


        #spatial connection probabilites on each subpopulation
        #Each key must correspond to a subpopulation like 'L23E' used everywhere else,
        #each array maps thalamic and intracortical connections.
        #First column is thalamic connections, and the rest intracortical,
        #ordered like 'L23E', 'L23I' etc., first row is normalised probability of
        #connection withing L1, L2, etc.;
        self.L_yXL = get_L_yXL(fname = self.connectivity_table,
                                         y = self.y,
                                         x_in_X = self.x_in_X,
                                         L = ['1','23','4','5','6'])
        
        #compute the cell type specificity
        self.T_yX = get_T_yX(fname=self.connectivity_table, y=self.y,
                             y_in_Y=self.y_in_Y, x_in_X=self.x_in_X,
                             F_y=self.F_y)
        
        Y, y = zip(*self.mapping_Yy)

        #assess relative distribution of synapses for a given celltype
        self.K_yXL = {}
        #self.T_yX = {}
        for i, (Y, y) in enumerate(self.mapping_Yy):
            #fill in K_yXL (layer specific connectivity)
            self.K_yXL[y] = (self.T_yX[i, ] * K_YX[np.array(self.Y)==Y, ] * self.L_yXL[y]).astype(int)
        
        #number of incoming connections per cell type per layer per cell 
        self.k_yXL = {}
        for y, N_y in zip(self.y, self.N_y):
            self.k_yXL.update({y : (1. * self.K_yXL[y]).astype(int) / N_y})

        #calculate corresponding connectivity to K_yXL
        self.C_yXL = {}
        for y, N_y in zip(self.y, self.N_y):
            self.C_yXL.update({y : 1. - (1.-1./(N_y* self.N_X))**self.K_yXL[y] })


################################################################################################################


class point_neuron_network_params(general_params):
    '''class point-neuron network parameters'''
    def __init__(self):
        '''class point-neuron network parameters'''

        # inherit general params
        general_params.__init__(self)
        

    ####################################
    #                                  #
    #                                  #
    #     SIMULATION PARAMETERS        #
    #                                  #
    #                                  #
    #################################### 

        # use same number of threads as MPI COMM.size()
        self.total_num_virtual_procs = SIZE

        ####################################
        # RNG PROPERTIES                   #
        #################################### 

        # offset for RNGs
        self.seed_offset = 45


        ####################################
        # RECORDING PARAMETERS             #        
        ####################################

        self.to_memory = False

        self.overwrite_existing_files = True 

        # recording can either be done from a fraction of neurons in each population or from a fixed number

        # whether to record spikes from a fixed fraction of neurons in each population. 
        self.record_fraction_neurons_spikes = True 

        if self.record_fraction_neurons_spikes:
            self.frac_rec_spikes = 1.
        else:
            self.n_rec_spikes = 100 

        # whether to record membrane potentials from a fixed fraction of neurons in each population
        self.record_fraction_neurons_voltage = False 

        if self.record_fraction_neurons_voltage:
            self.frac_rec_voltage = 0.1 
        else:
            self.n_rec_voltage = 100 

        # whether to record weighted input spikes from a fixed fraction of neurons in each population
        self.record_fraction_neurons_input_spikes = False

        if self.record_fraction_neurons_input_spikes:
            self.frac_rec_input_spikes = 0.1 
        else:
            self.n_rec_input_spikes = 0 

        # number of recorded neurons for depth resolved input currents
        self.n_rec_depth_resolved_input = 0 
 
        # whether to write any recorded cortical spikes to file
        self.save_cortical_spikes = True 

        # whether to write any recorded membrane potentials to file
        self.save_voltages = False 

        # whether to record thalamic spikes
        self.record_thalamic_spikes = True 

        # whether to write any recorded thalamic spikes to file
        self.save_thalamic_spikes = True 

        # global ID file name
        self.GID_filename = 'population_GIDs.dat'

        # readout global ID file name
        self.readout_GID_filename = 'readout_GIDs.dat' 

        # stem for spike detector file labels
        self.spike_detector_label = 'spikes_'

        # stem for voltmeter file labels
        self.voltmeter_label = 'voltages_'

        # stem for thalamic spike detector file labels
        self.th_spike_detector_label = 'spikes_0'

        # stem for in-degree file labels
        self.in_degree_label = 'in_degrees_'

        # stem for file labels for in-degree from thalamus 
        self.th_in_degree_label = 'in_degrees_th_'

        # stem for weighted input spikes labels
        self.weighted_input_spikes_label = 'weighted_input_spikes_'


    ####################################
    #                                  #
    #                                  #
    #        MODEL PARAMETERS          #
    #                                  #
    #                                  #
    ####################################


        ####################################
        # SCALING                          #
        ####################################  
       
        # scaling parameter for population sizes
        self.area = 1.
        
        # preserve indegrees when downscaling
        self.preserve_K = False         
        

        ####################################
        # SINGLE NEURON PARAMS             #
        ####################################

        # neuron model
        self.neuron_model = '/iaf_psc_exp'      

        # mean of initial membrane potential (mV)
        self.Vm0_mean = -58.0

        # std of initial membrane potential (mV)
        self.Vm0_std = 10.0

        # mean of threshold potential (mV)
        self.V_th_mean = -50.

        # std of threshold potential (mV)
        self.V_th_std = 0.
        
        self.model_params = { 'tau_m': 10.,        # membrane time constant (ms)
                              'tau_syn_ex': 0.5,   # excitatory synaptic time constant (ms)
                              'tau_syn_in': 0.5,   # inhibitory synaptic time constant (ms)
                              't_ref': 2.,         # absolute refractory period (ms)
                              'E_L': -65.,         # resting membrane potential (mV)
                              'V_th': self.V_th_mean, # spike threshold (mV)
                              'C_m': 250.,         # membrane capacitance (pF)
                              'V_reset': -65.      # reset potential (mV)
                              } 
        


        ####################################
        # EXTERNAL INPUTS                  #
        ####################################

        #number of external inputs (Potjans-Diesmann model 2012)
        self.K_bg = [[1600,    # layer 23 e
                      1500],   # layer 23 i
                     [2100,    # layer 4 e
                      1900],   # layer 4 i
                     [2000,    # layer 5 e
                      1900],   # layer 5 i
                     [2900,    # layer 6 e
                      2100]]   # layer 6 i
        
        # rate of Poisson input at each external input synapse (spikess)
        self.bg_rate = 0.       

        # rate of equivalent input used for DC amplitude calculation,
        # set to zero if self.bg_rate > 0.
        self.bg_rate_dc = 8.

        # DC amplitude at each external input synapse (pA)  
        # to each neuron via 'dc_amplitude  = tau_syn_ex/1000*bg_rate*PSC_ext'
        self.dc_amplitude = self.model_params["tau_syn_ex"] * self.bg_rate_dc *\
                            self._compute_J()

        # mean EPSP amplitude (mV) for thalamic and non-thalamic external input spikes
        self.PSP_ext = 0.15 	       

        # mean delay of thalamic input (ms)
        self.delay_th = 1.5  	

        # standard deviation relative to mean delay of thalamic input
        self.delay_th_rel_sd = 0.5   


        ####################################
        # THALAMIC INPUT VERSIONS  	   #
        ####################################
  
        # off-option for start of thalamic input versions
        self.off = 100.*self.tstop

   
        ## poisson_generator (pure Poisson input)
        self.th_poisson_start = self.off  	# onset (ms)
        self.th_poisson_duration = 10.	        # duration (ms)
        self.th_poisson_rate = 120. 	        # rate (spikess)


        ## spike_generator 
        # Note: This can be used with a large Gaussian delay distribution in order to mimic a 
        #       Gaussian pulse packet which is different for each thalamic neuron
        self.th_spike_times = [self.off]	# time of the thalamic pulses (ms)


        ## sinusoidal_poisson_generator (oscillatory Poisson input)
        self.th_sin_start = self.off      	# onset (ms)
        self.th_sin_duration = 5000.  	        # duration (ms)
        self.th_sin_mean_rate = 30. 	        # mean rate (spikess)
        self.th_sin_fluc_rate = 30.  	        # rate modulation amplitude (spikess)
        self.th_sin_freq = 15. 	                # frequency of the rate modulation (Hz)
        self.th_sin_phase = 0.                  # phase of rate modulation (deg)


        ## Gaussian_pulse_packages
        self.th_gauss_times = [self.off]                # package center times
        self.th_gauss_num_spikes_per_packet = 1 	# number of spikes per packet
        self.th_gauss_sd = 5. 				# std of Gaussian pulse packet (ms^2)


        ####################################
        # SPATIAL ORGANIZATION             #
        ####################################
        
        # needed for spatially resolved input currents
        
        # number of layers TODO: find a better solution for that
        self.num_input_layers = 5


    def _compute_J(self):
        '''
        Compute the current amplitude corresponding to the exponential
        synapse model PSP amplitude
        
        Derivation using sympy:
        ::
            from sympy import *
            #define symbols
            t, tm, Cm, ts, Is, Vmax = symbols('t tm Cm ts Is Vmax')
            
            #assume zero delay, t >= 0
            #using eq. 8.10 in Sterrat et al
            V = tm*ts*Is*(exp(-t/tm) - exp(-t/ts)) / (tm-ts) / Cm
            print 'V = %s' % V
            
            #find time of V == Vmax
            dVdt = diff(V, t)
            print 'dVdt = %s' % dVdt
            
            [t] = solve(dVdt, t)
            print 't(t@dVdT==Vmax) = %s' % t
            
            #solve for Is at time of maxima
            V = tm*ts*Is*(exp(-t/tm) - exp(-t/ts)) / (tm-ts) / Cm
            print 'V(%s) = %s' % (t, V)
            
            [Is] = solve(V-Vmax, Is)
            print 'Is = %s' % Is
        
        resulting in:
        ::
            Cm*Vmax*(-tm + ts)/(tm*ts*(exp(tm*log(ts/tm)/(tm - ts))
                                     - exp(ts*log(ts/tm)/(tm - ts))))
        
        '''
        #LIF params
        tm = self.model_params['tau_m']
        Cm = self.model_params['C_m']
        
        #synapse
        ts = self.model_params['tau_syn_ex']
        Vmax = self.PSP_e
        
        #max current amplitude
        J = Cm*Vmax*(-tm + ts)/(tm*ts*(np.exp(tm*np.log(ts/tm)/(tm - ts))
                                     - np.exp(ts*np.log(ts/tm)/(tm - ts))))
        
        #unit conversion pF*mV -> nA
        J *= 1E-3
        
        return J
 

class multicompartment_params(point_neuron_network_params):
    '''
    Inherited class defining additional attributes needed by e.g., the
    classes population.Population and population.DummyNetwork

    This class do not take any kwargs    

    '''
    def __init__(self):
        '''
        Inherited class defining additional attributes needed by e.g., the
        classes population.Population and population.DummyNetwork
        
        This class do not take any kwargs    
        
        '''
        
        # initialize parent classes
        point_neuron_network_params.__init__(self)
 

    ####################################
    #                                  #
    #                                  #
    #     SIMULATION PARAMETERS        #
    #                                  #
    #                                  #
    ####################################
       

        #######################################
        # PARAMETERS FOR LOADING NEST RESULTS #
        #######################################


        # parameters for class population.DummyNetwork class
        self.networkSimParams = {
            'simtime' :     self.tstop - self.tstart,
            'dt' :          self.dt,
            'spike_output_path' : self.spike_output_path,
            'label' :       'population_spikes',
            'ext' :         'gdf',
            'GIDs' : self.get_GIDs(),
        }


        # Switch for current source density computations
        self.calculateCSD = True


    ####################################
    #                                  #
    #                                  #
    #        MODEL PARAMETERS          #
    #                                  #
    #                                  #
    ####################################

       
        ####################################
        # SCALING (VOLUME not density)     #
        ####################################  
           
        self.SCALING = 1.
        
  
        ####################################
        # MORPHOLOGIES                     #
        ####################################

        # list of morphology files with default location, testing = True
        # will point to simplified morphologies
        testing = True
        if testing:
            self.PATH_m_y = os.path.join('morphologies', 'ballnsticks')
            self.m_y = [Y + '_' + y + '.hoc' for Y, y in self.mapping_Yy]
        else:
            self.PATH_m_y = os.path.join('morphologies', 'stretched')
            self.m_y = [
                'L23E_oi24rpy1.hoc',
                'L23I_oi38lbc1.hoc',
                'L23I_oi38lbc1.hoc',
            
                'L4E_53rpy1.hoc',
                'L4E_j7_L4stellate.hoc',
                'L4E_j7_L4stellate.hoc',
                'L4I_oi26rbc1.hoc',
                'L4I_oi26rbc1.hoc',
            
                'L5E_oi15rpy4.hoc',
                'L5E_j4a.hoc',
                'L5I_oi15rbc1.hoc',
                'L5I_oi15rbc1.hoc',
            
                'L6E_51-2a.CNG.hoc',
                'L6E_oi15rpy4.hoc',
                'L6I_oi15rbc1.hoc',
                'L6I_oi15rbc1.hoc',
                ]


        ####################################
        # CONNECTION WEIGHTS               #
        ####################################
        
        # compute the synapse weight from fundamentals of exp synapse LIF neuron
        self.J = self._compute_J()
        
        # set up matrix containing the synapse weights between any population X
        # and population Y, including exceptions for certain connections
        J_YX = np.zeros(self.C_YX.shape)
        J_YX += self.J
        J_YX[:, 2::2] *= self.g
        if hasattr(self, 'PSP_23e_4e'):
            J_YX[0, 3] *= self.PSP_23e_4e / self.PSP_e
        if hasattr(self, 'g_4e_4i'):
            J_YX[2, 4] *= self.g_4e_4i / self.g
        

        # extrapolate weights between populations X and
        # cell type y in population Y
        self.J_yX = {}
        for Y, y in self.mapping_Yy:
            [i] = np.where(np.array(self.Y) == Y)[0]
            self.J_yX.update({y : J_YX[i, ]})
        
    
        ####################################
        # GEOMETRY OF CORTICAL COLUMN      #
        ####################################
        
        # set the boundaries of each layer, L1->L6,
        # and mean depth of soma layers
        self.layerBoundaries = np.array([[     0.0,   -81.6],
                                          [  -81.6,  -587.1],
                                          [ -587.1,  -922.2],
                                          [ -922.2, -1170.0],
                                          [-1170.0, -1491.7]])
        
        # assess depth of each 16 subpopulation
        self.depths = self._calcDepths()
        
        # make a nice structure with data for each subpopulation
        self.y_zip_list = zip(self.y, self.m_y,
                            self.depths, self.N_y)



        ##############################################################
        # POPULATION PARAMS (cells, population, synapses, electrode) #
        ##############################################################


        # Global LFPy.Cell-parameters, by default shared between populations
        # Some passive parameters will not be fully consistent with LIF params
        self.cellParams = {
            'v_init' : self.model_params['E_L'],
            'passive' : True,
            'rm' : self.model_params['tau_m'] * 1E3 / 1.0, #assyme cm=1
            'cm' : 1.0,
            'Ra' : 150,
            'e_pas' : self.model_params['E_L'],    
            'nsegs_method' : 'lambda_f',
            'lambda_f' : 100,
            'timeres_NEURON' : self.dt,
            'timeres_python' : self.dt,
            'tstartms' : self.tstart,
            'tstopms' : self.tstop,
            'verbose' : False,
        }
        

        # layer specific LFPy.Cell-parameters as nested dictionary
        self.yCellParams = self._yCellParams()
        
        
        # set the axis of which each cell type y is randomly rotated,
        # SS types and INs are rotated around both x- and z-axis
        # in the population class, while P-types are
        # only rotated around the z-axis
        self.rand_rot_axis = {}
        for y, _, _, _ in self.y_zip_list:
            #identify pyramidal cell populations:
            if y.rfind('p') >= 0:
                self.rand_rot_axis.update({y : ['z']})
            else:
                self.rand_rot_axis.update({y : ['x', 'z']})
        
        
        # additional simulation kwargs, see LFPy.Cell.simulate() docstring
        self.simulationParams = {}
        
                
        # a dict setting the number of cells N_y and geometry
        # of cell type population y
        self.populationParams = {}
        for y, _, depth, N_y in self.y_zip_list:
            self.populationParams.update({
                y : {
                    #'number' : 4 if 'p' in y else 1,
                    'number' : int(N_y*self.SCALING),
                    'radius' : np.sqrt(1000**2 / np.pi),
                    'z_min' : depth - 25,
                    'z_max' : depth + 25,
                    'min_cell_interdist' : 1.,            
                }
            })

        # Set up cell type specific synapse parameters in terms of synapse model
        # and synapse locations
        self.synParams = {}
        for y in self.y:
            if y.rfind('p') >= 0:
                #pyramidal types have apical dendrites
                section = ['apic', 'dend']
            else:
                #other cell types do not
                section = ['dend']

            self.synParams.update({
                y : {
                    'syntype' : 'ExpSynI',  #current based exponential synapse
                    'section' : section,
                    'tau' : self.model_params["tau_syn_ex"],
                },
            })


        #synaptic delay parameters, loc and scale is mean and std for every
        #network population, negative values will be removed
        self.synDelayLoc, self.synDelayScale = self._synDelayParams()

         
        # Define electrode geometry corresponding to a laminar electrode,
        # where contact points have a radius r, surface normal vectors N,
        # and LFP calculated as the average LFP in n random points on
        # each contact. Recording electrode emulate NeuroNexus array,
        # contact 0 is superficial
        self.electrodeParams = {
            #contact locations:
            'x' : np.zeros(16),
            'y' : np.zeros(16),
            'z' : -np.mgrid[0:16] * 100,
            #extracellular conductivity:
            'sigma' : 0.3,
            #contact surface normals, radius, n-point averaging
            'N' : np.array([[1, 0, 0]]*16),
            'r' : 7.5,
            'n' : 50,
            'seedvalue' : None,
            #dendrite line sources, soma sphere source (Linden2014)
            'method' : 'som_as_point',
            #no somas within the constraints of the "electrode shank":
            'r_z': np.array([[-1E199, -1600, -1550, 1E99],[0, 0, 10, 10]]),
        }
        
        
        #these variables will be saved to file for each cell and electrdoe object
        self.savelist = [
            'somav',
            'timeres_NEURON',
            'timeres_python',
            'somapos',
            'x',
            'y',
            'z',
            'LFP',
            'CSD',
            'morphology',
            'default_rotation',
            'electrodecoeff',
        ]
        
        
        #########################################
        # MISC                                  #
        #########################################
        
        #time resolution of downsampled data in ms
        self.dt_output = 1.
        
        #set fraction of neurons from population which LFP output is stored
        self.recordSingleContribFrac = 0.


    def get_GIDs(self):
        GIDs = {}
        ind = 1
        for i, (X, N_X) in enumerate(zip(self.X, self.N_X)):
            GIDs[X] = [ind, N_X]
            ind += N_X
        return GIDs


    def _synDelayParams(self):
        '''
        set up the detailed synaptic delay parameters,
        loc is mean delay,
        scale is std with low bound cutoff,
        assumes numpy.random.normal is used later
        '''
        delays = {}
        #mean delays
        loc = np.zeros((len(self.y), len(self.X)))
        loc[:, 0] = self.delays[0]
        loc[:, 1::2] = self.delays[0]
        loc[:, 2::2] = self.delays[1]
        #standard deviations
        scale = loc * self.delay_rel_sd
        
        #prepare output
        delay_loc = {}
        for i, y in enumerate(self.y):
            delay_loc.update({y : loc[i]})
        
        delay_scale = {}
        for i, y in enumerate(self.y):
            delay_scale.update({y : scale[i]})
                
        return delay_loc, delay_scale


    def _calcDepths(self):
        '''
        return the cortical depth of each subpopulation
        '''
        depths = self.layerBoundaries.mean(axis=1)[1:]

        depth_y = []
        for y in self.y:
            if y in ['p23', 'b23', 'nb23']:
                depth_y = np.r_[depth_y, depths[0]]
            elif y in ['p4', 'ss4(L23)', 'ss4(L4)', 'b4', 'nb4']:
                depth_y = np.r_[depth_y, depths[1]]
            elif y in ['p5(L23)', 'p5(L56)', 'b5', 'nb5']:
                depth_y = np.r_[depth_y, depths[2]]
            elif y in ['p6(L4)', 'p6(L56)', 'b6', 'nb6']:
                depth_y = np.r_[depth_y, depths[3]]
            else:
                raise Exception, 'this aint right'
                
        return depth_y


    def _yCellParams(self):
        '''
        Return dict with parameters for each population.
        The main operation is filling in cell type specific morphology
        '''
        #cell type specific parameters going into LFPy.Cell        
        yCellParams = {}
        for layer, morpho, _, _ in self.y_zip_list:
            yCellParams.update({layer : self.cellParams.copy()})
            yCellParams[layer].update({
                'morphology' : os.path.join(self.PATH_m_y, morpho),
            })
        return yCellParams
 
        
if __name__ == '__main__':
    params = multicompartment_params()
     
    print dir(params)



#!/usr/bin/env python
'''
**hybridLFPy**
==============

Provides methods for estimating extracellular potentials of simplified spiking
neuron network models.


How to use the documentation
----------------------------

Documentation is available in two forms:
    1.  Docstrings provided with the code, e.g., as hybridLFPy? within IPython
    2.  Autogenerated sphinx-built output, compiled issuing:
        ::
            
            sphinx-build -b html docs documentation

        in the root folder of the package sources


Available classes
-----------------
``CachedNetwork``
    Offline interface between network spike events and used by class Population

``CachedNoiseNetwork``
    Creation of Poisson spiketrains of a putative network model, interfaces
    class Population
    
``CachedFixedSpikesNetwork``
    Creation of identical spiketrains per population of a putative network
    model, interface to class Population

``GDF``
    Class using sqlite to efficiently store and enquire large amounts of spike
    output data, used by Cached*Network

``PopulationSuper``
    Parent class setting up a base population of multicompartment neurons

``Population``
    Daughter of PopulationSuper, using CachedNetwork spike events as synapse
    activation times with layer and cell type connection specificity

``PostProcess``
    Methods for processing output of multiple instances of class Population


Available utilities
-------------------
``csd``
    Ground truth current-source density estimation from multicompartment models

``helpers``
    Various methods used throughout simulations

``setup_file_dest``
    Setup destination folders of simulation output files

``test``
    Run a series of unit tests
'''

from .cachednetworks import CachedNetwork, CachedNoiseNetwork, CachedFixedSpikesNetwork
from .gdf import GDF
from .population import PopulationSuper, Population
from .postproc import PostProcess
from . import csd
from . import helpers
from .helpers import setup_file_dest
from .testing import test



#!/usr/bin/env python
import numpy as np
import h5py
import os
import glob
import tarfile
from warnings import warn
from mpi4py import MPI

################# Initialization of MPI stuff ##################################
COMM = MPI.COMM_WORLD
SIZE = COMM.Get_size()
RANK = COMM.Get_rank()

class PostProcess(object):
    """
    class `PostProcess`: Methods to deal with the contributions of every
    postsynaptic sub-population.


    Parameters
    ----------
    y : list
        Postsynaptic cell-type or population-names.
    dt_output : float
        Time resolution of output data.
    savefolder : str
        Path to main output folder.
    mapping_Yy : list
        List of tuples, each tuple pairing population with cell type, e.g., [('L4E', 'p4'), ('L4E', 'ss4')].
    cells_subfolder : str
        Folder under `savefolder` containing cell output.
    populations_subfolder : str
        Folder under `savefolder` containing population specific output.
    figures_subfolder : str
        Folder under `savefolder` containing figs.

    """
    def __init__(self,
                 y=['EX', 'IN'],
                 dt_output=1.,
                 mapping_Yy=[('EX', 'EX'),
                               ('IN', 'IN')],
                 savelist=['somav', 'somapos', 'x', 'y', 'z', 'LFP', 'CSD'],
                 savefolder='simulation_output_example_brunel',
                 cells_subfolder='cells',
                 populations_subfolder='populations',
                 figures_subfolder='figures',
                 output_file='{}_population_{}',
                 compound_file='{}sum.h5',
                 ):
        """
        class `PostProcess`: Methods to deal with the contributions of every
        postsynaptic sub-population.

    
        Parameters
        ----------
        y : list
            List of postsynaptic cell-type or population-names.
        dt_output : float
            Time resolution of output data.
        mapping_Yy : list
            List of tuples, each tuple pairing population with cell type,
            e.g., [('L4E', 'p4'), ('L4E', 'ss4')].
        savelist : list
            List of strings, each corresponding to LFPy.Cell attributes
        savefolder : str
            Path to main output folder.
        cells_subfolder : str
            Folder under `savefolder` containing cell output.
        populations_subfolder : str
            Folder under `savefolder` containing population specific output.
        figures_subfolder : str
            Folder under `savefolder` containing figs.
        output_file : str
            formattable file name for population signals, e.g.,
            '{}_population_{}.h5'
        compound_file : str
            formattable file name for population signals, e.g.,
            '{}sum.h5'
        
    
        """
        #set some attributes
        self.y = y
        self.dt_output = dt_output
        self.mapping_Yy = mapping_Yy
        self.savelist = savelist
        self.savefolder = savefolder
        self.cells_path = os.path.join(savefolder, cells_subfolder)
        self.populations_path = os.path.join(savefolder, populations_subfolder)
        self.figures_path = os.path.join(savefolder, figures_subfolder)
        self.output_file = output_file
        self.compound_file = compound_file
        
        #set up subfolders
        if RANK == 0:
            self._set_up_savefolder()
        else:
            pass


    def run(self):
        """ Perform the postprocessing steps, computing compound signals from
        cell-specific output files.
        """
        if RANK == 0:
            if 'LFP' in self.savelist:
                #get the per population LFPs and total LFP from all populations:
                self.LFPdict, self.LFPsum = self.calc_lfp()
                self.LFPdictLayer = self.calc_lfp_layer()
    
                #save global LFP sum, and from L23E, L4I etc.:
                f = h5py.File(os.path.join(self.savefolder,
                                           self.compound_file.format('LFP')
                                           ), 'w')
                f['srate'] = 1E3 / self.dt_output
                f.create_dataset('data', data=self.LFPsum, compression=4)
                f.close()
    
                for key, value in list(self.LFPdictLayer.items()):
                    f = h5py.File(os.path.join(self.populations_path,
                                               self.output_file.format(key,
                                                                       'LFP.h5')
                                               ), 'w')
                    f['srate'] = 1E3 / self.dt_output
                    f.create_dataset('data', data=value, compression=4)
                    f.close()

            if 'CSD' in self.savelist:
                #get the per population CSDs and total CSD from all populations:
                self.CSDdict, self.CSDsum = self.calc_csd()
                self.CSDdictLayer = self.calc_csd_layer()
    
                #save global CSD sum, and from L23E, L4I etc.:
                f = h5py.File(os.path.join(self.savefolder,
                                           self.compound_file.format('CSD')),
                              'w')
                f['srate'] = 1E3 / self.dt_output
                f.create_dataset('data', data=self.CSDsum, compression=4)
                f.close()
    
                for key, value in list(self.CSDdictLayer.items()):
                    f = h5py.File(os.path.join(self.populations_path,
                                               self.output_file.format(key,
                                                                       'CSD.h5')
                                               ), 'w')
                    f['srate'] = 1E3 / self.dt_output
                    f.create_dataset('data', data=value, compression=4)
                    f.close()

        else:
            pass

        ##collect matrices with the single cell contributions in parallel
        #self.collectSingleContribs()


    def _set_up_savefolder(self):
        """ Create catalogs for different file output to clean up savefolder.
        """
        if not os.path.isdir(self.cells_path):
            os.mkdir(self.cells_path)

        if not os.path.isdir(self.figures_path):
            os.mkdir(self.figures_path)

        if not os.path.isdir(self.populations_path):
            os.mkdir(self.populations_path)


    def calc_lfp(self):
        """ Sum all the LFP contributions from every cell type.
        """

        LFParray = np.array([])
        LFPdict = {}

        i = 0
        for y in self.y:
            fil = os.path.join(self.populations_path,
                               self.output_file.format(y, 'LFP.h5'))

            f = h5py.File(fil)

            if i == 0:
                LFParray = np.zeros((len(self.y),
                                    f['data'].shape[0], f['data'].shape[1]))

            #fill in
            LFParray[i, ] = f['data'].value

            LFPdict.update({y : f['data'].value})

            f.close()

            i += 1

        return LFPdict,  LFParray.sum(axis=0)


    def calc_csd(self):
        """ Sum all the CSD contributions from every layer.
        """

        CSDarray = np.array([])
        CSDdict = {}

        i = 0
        for y in self.y:
            fil = os.path.join(self.populations_path,
                               self.output_file.format(y, 'CSD.h5'))

            f = h5py.File(fil)

            if i == 0:
                CSDarray = np.zeros((len(self.y),
                                    f['data'].shape[0], f['data'].shape[1]))

            #fill in
            CSDarray[i, ] = f['data'].value

            CSDdict.update({y : f['data'].value})

            f.close()

            i += 1

        return CSDdict,  CSDarray.sum(axis=0)


    def calc_lfp_layer(self):
        """
        Calculate the LFP from concatenated subpopulations residing in a
        certain layer, e.g all L4E pops are summed, according to the `mapping_Yy`
        attribute of the `hybridLFPy.Population` objects.
        """
        LFPdict = {}

        lastY = None
        for Y, y in self.mapping_Yy:
            if lastY != Y:
                try:
                    LFPdict.update({Y : self.LFPdict[y]})
                except KeyError:
                    pass
            else:
                try:
                    LFPdict[Y] += self.LFPdict[y]
                except KeyError:
                    pass
            lastY = Y

        return LFPdict


    def calc_csd_layer(self):
        """
        Calculate the CSD from concatenated subpopulations residing in a
        certain layer, e.g all L4E pops are summed, according to the `mapping_Yy`
        attribute of the `hybridLFPy.Population` objects.
        """
        CSDdict = {}

        lastY = None
        for Y, y in self.mapping_Yy:
            if lastY != Y:
                try:
                    CSDdict.update({Y : self.CSDdict[y]})
                except KeyError:
                    pass
            else:
                try:
                    CSDdict[Y] += self.CSDdict[y]
                except KeyError:
                    pass
            lastY = Y

        return CSDdict


    def create_tar_archive(self):
        """ Create a tar archive of the main simulation outputs.
        """
        #file filter
        EXCLUDE_FILES = glob.glob(os.path.join(self.savefolder, 'cells'))
        EXCLUDE_FILES += glob.glob(os.path.join(self.savefolder,
                                                'populations', 'subsamples'))
        EXCLUDE_FILES += glob.glob(os.path.join(self.savefolder,
                                                'raw_nest_output'))

        def filter_function(tarinfo):
            print(tarinfo.name)
            if len([f for f in EXCLUDE_FILES if os.path.split(tarinfo.name)[-1]
                    in os.path.split(f)[-1]]) > 0 or \
               len([f for f in EXCLUDE_FILES if os.path.split(tarinfo.path)[-1]
                    in os.path.split(f)[-1]]) > 0:
                print('excluding %s' % tarinfo.name)

                return None
            else:
                return tarinfo

        if RANK == 0:
            print('creating archive %s' % (self.savefolder + '.tar'))
            #open file
            f = tarfile.open(self.savefolder + '.tar', 'w')
            #avoid adding files to repo as /scratch/$USER/hybrid_model/...
            arcname = os.path.split(self.savefolder)[-1]

            f.add(name=self.savefolder,
                  arcname=arcname,
                  filter=filter_function)
            f.close()

        #resync
        COMM.Barrier()




#!/usr/bin/env python
'''setup file for the hybridLFPy python module'''
import os
from setuptools import setup

with open('README.md') as file:
    long_description = file.read()

setup(
    name = 'hybridLFPy',
    version = '0.1.2',
    maintainer = 'Espen Hagen',
    maintainer_email = 'e.hagen@fz-juelich.de',
    url = 'https://github.com/INM-6/hybridLFPy',
    download_url = 'https://github.com/INM-6/hybridLFPy/tarball/v0.1.2',
    packages = ['hybridLFPy'],
    provides = ['hybridLFPy'],
    package_data = {'hybridLFPy' : [os.path.join('testing', 'testing-X-0.gdf')]},
    description = 'methods to calculate LFPs with spike events from network sim',
    long_description = long_description,
    license='LICENSE',
        classifiers=[
            'License :: OSI Approved :: GNU General Public License (GPL)',
            'Programming Language :: Python',
            'Programming Language :: Python :: 2.6',
            'Programming Language :: Python :: 2.7',
            'Programming Language :: Python :: 3.3',
            'Programming Language :: Cython',
            'Operating System :: OS Independent',
            'Topic :: Scientific/Engineering',
            'Topic :: Scientific/Engineering :: Physics',
            'Intended Audience :: Developers',
            'Intended Audience :: Science/Research',
            'Development Status :: 4 - Beta',
            ],
    install_requires = [
        'numpy', 'scipy', 'matplotlib', 'neuron', 'LFPy', 'mpi4py',
        'NeuroTools',
        ],

)



#!/usr/bin/env python
"""
Documentation:

This is a script containing general helper functions which can be applied
to specialized cases.
"""

import numpy as np
import os
import stat
import shutil
import glob
import copy
import scipy.signal as ss
import h5py
from mpi4py import MPI


###################################
# Initialization of MPI stuff     #
###################################
COMM = MPI.COMM_WORLD
SIZE = COMM.Get_size()
RANK = COMM.Get_rank()



#######################################
### DATA I/O                        ###
#######################################


def read_gdf(fname):
    """
    Fast line-by-line gdf-file reader.
    
    
    Parameters
    ----------
    fname : str 
        Path to gdf-file.
    
    
    Returns
    -------
    numpy.ndarray
        ([gid, val0, val1, **]), dtype=object) mixed datatype array
            
    """
    
    gdf_file = open(fname, 'r')
    gdf = []
    for l in gdf_file:
        data = l.split()
        gdf += [data]

    gdf = np.array(gdf, dtype=object)
    
    if gdf.size > 0:
        gdf[:, 0] = gdf[:, 0].astype(int)
        gdf[:, 1:] = gdf[:, 1:].astype(float)
    
    return np.array(gdf)


def write_gdf(gdf, fname):
    """
    Fast line-by-line gdf-file write function
    
    
    Parameters
    ----------
    gdf : numpy.ndarray
        Column 0 is gids, columns 1: are values.
    fname : str
        Path to gdf-file.
    
    
    Returns
    -------
    None
    
    """
    gdf_file = open(fname, 'w')
    for line in gdf:
        for i in np.arange(len(line)):
            gdf_file.write(str(line[i]) + '\t')
        gdf_file.write('\n')
    
    return None


def load_h5_data(path='', data_type='LFP', y=None, electrode=None,
                 warmup=0., scaling=1.):
    """
    Function loading results from hdf5 file
    
    
    Parameters
    ----------
    path : str
        Path to hdf5-file
    data_type : str
        Signal types in ['CSD' , 'LFP', 'CSDsum', 'LFPsum'].
    y : None or str
        Name of population.
    electrode : None or int
        TODO: update, electrode is NOT USED
    warmup : float
        Lower cutoff of time series to remove possible transients
    scaling : float,
        Scaling factor for population size that determines the amount of loaded
        single-cell signals
                
    
    Returns
    ----------
    numpy.ndarray
        [electrode id, compound signal] if `y` is None
    numpy.ndarray
        [cell id, electrode, single-cell signal] otherwise
    
    """
    assert y is not None or electrode is not None
    
    if y is not None:
        f = h5py.File(os.path.join(path, '%s_%ss.h5' %(y,data_type)))
        data = f['data'].value[:,:, warmup:]
        if scaling != 1.:
            np.random.shuffle(data)
            num_cells = int(len(data)*scaling)
            data = data[:num_cells,:, warmup:]
    else:
        f = h5py.File(os.path.join(path, '%ssum.h5' %data_type))
        data = f['data'].value[:, warmup:]

    return data


def dump_dict_of_nested_lists_to_h5(fname, data):
    """
    Take nested list structure and dump it in hdf5 file.


    Parameters
    ----------
    fname : str 
        Filename
    data : dict(list(numpy.ndarray))
        Dict of nested lists with variable len arrays.
    
    
    Returns
    -------
    None

    """
    # Open file
    print('writing to file: %s' % fname)
    f = h5py.File(fname)
    # Iterate over values
    for i, ivalue in list(data.items()):
        igrp = f.create_group(str(i))
        for j, jvalue in enumerate(ivalue):
            jgrp = igrp.create_group(str(j))
            for k, kvalue in enumerate(jvalue):
                if kvalue.size > 0:
                    dset = jgrp.create_dataset(str(k), data=kvalue,
                                               compression='gzip')
                else:
                    dset = jgrp.create_dataset(str(k), data=kvalue,
                                               maxshape=(None, ),
                                               compression='gzip')
    # Close file
    f.close()


def load_dict_of_nested_lists_from_h5(fname, toplevelkeys=None):
    """
    Load nested list structure from hdf5 file


    Parameters
    ----------        
    fname : str
        Filename
    toplevelkeys : None or iterable, 
        Load a two(default) or three-layered structure.


    Returns
    -------
    dict(list(numpy.ndarray))
        dictionary of nested lists with variable length array data.
    
    """
    
    # Container:
    data = {}

    # Open file object
    f = h5py.File(fname, 'r')

    # Iterate over partial dataset
    if toplevelkeys is not None:
        for i in toplevelkeys:
            ivalue = f[str(i)]
            data[i] = []
            for j, jvalue in enumerate(ivalue.values()):
                data[int(i)].append([])
                for k, kvalue in enumerate(jvalue.values()):
                    data[i][j].append(kvalue.value)
    else:
        for i, ivalue in list(f.items()):
            i = int(i)
            data[i] = []
            for j, jvalue in enumerate(ivalue.values()):
                data[i].append([])
                for k, kvalue in enumerate(jvalue.values()):
                    data[i][j].append(kvalue.value)

    # Close dataset
    f.close()

    return data


def setup_file_dest(params, clearDestination=True):
    """
    Function to set up the file catalog structure for simulation output
    
    
    Parameters
    ----------  
    params : object 
        e.g., `cellsim16popsParams.multicompartment_params()`
    clear_dest : bool 
        Savefolder will be cleared if already existing.
    
    
    Returns
    -------
    None
    
    """
    if RANK == 0:
        if not os.path.isdir(params.savefolder):
            os.mkdir(params.savefolder)
            assert(os.path.isdir(params.savefolder))
        else:
            if clearDestination:
                print('removing folder tree %s' % params.savefolder)
                while os.path.isdir(params.savefolder):
                    try:
                        os.system('find %s -delete' % params.savefolder)
                    except:
                        shutil.rmtree(params.savefolder)
                os.mkdir(params.savefolder)
                assert(os.path.isdir(params.savefolder))
        
        if not os.path.isdir(params.sim_scripts_path):
            print('creating %s' % params.sim_scripts_path)
            os.mkdir(params.sim_scripts_path)
        
        if not os.path.isdir(params.cells_path):
            print('creating %s' % params.cells_path)
            os.mkdir(params.cells_path)
        
        if not os.path.isdir(params.figures_path):
            print('creating %s' % params.figures_path)
            os.mkdir(params.figures_path)
        
        if not os.path.isdir(params.populations_path):
            print('creating %s' % params.populations_path)
            os.mkdir(params.populations_path)
        
        try:
            if not os.path.isdir(params.raw_nest_output_path):
                print('creating %s' % params.raw_nest_output_path)
                os.mkdir(params.raw_nest_output_path)
        except:
            pass
        
        if not os.path.isdir(params.spike_output_path):
            print('creating %s' % params.spike_output_path)
            os.mkdir(params.spike_output_path)
    
        for f in ['cellsim16popsParams.py',
                  'cellsim16pops.py',
                  'example_brunel.py',
                  'brunel_alpha_nest.py',
                  'mesocircuit.sli',
                  'mesocircuit_LFP_model.py',
                  'binzegger_connectivity_table.json', 
                  'nest_simulation.py',
                  'microcircuit.sli']:
            if os.path.isfile(f):
                if not os.path.exists(os.path.join(params.sim_scripts_path, f)):
                    shutil.copy(f, os.path.join(params.sim_scripts_path, f))
                    os.chmod(os.path.join(params.sim_scripts_path, f),
                             stat.S_IREAD)
       
    COMM.Barrier()
    

#######################################
### GENERAL                         ###
#######################################


def calculate_fft(data, tbin):
    """
    Function to calculate the Fourier transform of data.
    
    
    Parameters
    ----------
    data : numpy.ndarray
        1D or 2D array containing time series.
    tbin : float
        Bin size of time series (in ms).
    
    
    Returns
    -------
    freqs : numpy.ndarray
        Frequency axis of signal in Fourier space.         
    fft : numpy.ndarray
        Signal in Fourier space.
        
    """
    if len(np.shape(data)) > 1:
        n = len(data[0])
        return np.fft.fftfreq(n, tbin * 1e-3), np.fft.fft(data, axis=1)
    else:
        n = len(data)
        return np.fft.fftfreq(n, tbin * 1e-3), np.fft.fft(data)


#######################################
### DATA MANIPULATION               ###
#######################################

def centralize(data, time=False, units=False):
    """
    Function to subtract the mean across time and/or across units from data
    
    
    Parameters
    ----------  
    data : numpy.ndarray
        1D or 2D array containing time series, 1st index: unit, 2nd index: time
    time : bool
        True: subtract mean across time.
    units : bool
        True: subtract mean across units.
            
    
    Returns
    -------
    numpy.ndarray
        1D or 0D array of centralized signal.
        
    """
    assert(time is not False or units is not False)
    res = copy.copy(data)
    
    if time is True:
        res = np.array([x - np.mean(x) for x in res])
    
    if units is True:
        res = np.array(res - np.mean(res, axis=0))
    
    return res


def normalize(data):
    """
    Function to normalize data to have mean 0 and unity standard deviation
    (also called z-transform)
    
    
    Parameters
    ----------
    data : numpy.ndarray
    
    
    Returns
    -------
    numpy.ndarray
        z-transform of input array
    
    """
    data = data.astype(float)
    data -= data.mean()
    
    return data / data.std()


#######################################
### FILTER                          ###
#######################################

def movav(y, Dx, dx):
    """
    Moving average rectangular window filter:
    calculate average of signal y by using sliding rectangular
    window of size Dx using binsize dx
    
    
    Parameters
    ----------
    y : numpy.ndarray
        Signal
    Dx : float
        Window length of filter.
    dx : float
        Bin size of signal sampling.
                
    
    Returns
    -------
    numpy.ndarray
        Filtered signal.
    
    """
    if Dx <= dx:
        return y
    else:
        ly = len(y)
        r = np.zeros(ly)
        n = np.int(np.round((Dx / dx)))
        r[0:np.int(n / 2.)] = 1.0 / n
        r[-np.int(n / 2.)::] = 1.0 / n
        R = np.fft.fft(r)
        Y = np.fft.fft(y)
        yf = np.fft.ifft(Y * R)
        return yf


def decimate(x, q=10, n=4, k=0.8, filterfun=ss.cheby1):
    """
    scipy.signal.decimate like downsampling using filtfilt instead of lfilter,
    and filter coeffs from butterworth or chebyshev type 1.


    Parameters
    ----------
    x : numpy.ndarray
        Array to be downsampled along last axis.
    q : int 
        Downsampling factor.
    n : int
        Butterworth filter order.
    k : float
        Aliasing filter critical frequency will be set as Wn=k/q.
    filterfun : function
        `scipy.signal.filter_design.cheby1` or
        `scipy.signal.filter_design.butter` function

    Returns
    -------
    numpy.ndarray
        Array of downsampled signal.
              
    """
    if not isinstance(q, int):
        raise TypeError("q must be an integer")

    if n is None:
        n = 1

    if filterfun == ss.butter:
        b, a = filterfun(n, k / q)
    elif filterfun == ss.cheby1:
        b, a = filterfun(n, 0.05, k / q)
    else:
        raise Exception('only ss.butter or ss.cheby1 supported')

    try:
        y = ss.filtfilt(b, a, x)
    except: # Multidim array can only be processed at once for scipy >= 0.9.0
        y = []
        for data in x:
            y.append(ss.filtfilt(b, a, data))
        y = np.array(y)

    try:
        return y[:, ::q]
    except:
        return y[::q]


#######################################
### CORRELATION ANALYSIS            ###
#######################################


def mean(data, units=False, time=False):
    """
    Function to compute mean of data


    Parameters
    ---------- 
    data : numpy.ndarray
        1st axis unit, 2nd axis time
    units : bool
        Average over units
    time : bool 
        Average over time


    Returns
    -------
    if units=False and time=False: 
        error
    if units=True: 
        1 dim numpy.ndarray; time series
    if time=True: 
        1 dim numpy.ndarray; series of unit means across time
    if units=True and time=True: 
        float; unit and time mean


    Examples
    --------
    >>> mean(np.array([[1, 2, 3], [4, 5, 6]]), units=True)
    array([ 2.5,  3.5,  4.5])

    >>> mean(np.array([[1, 2, 3], [4, 5, 6]]), time=True)
    array([ 2.,  5.])

    >>> mean(np.array([[1, 2, 3], [4, 5, 6]]), units=True,time=True)
    3.5

    """

    assert(units is not False or time is not False)
    if units is True and time is False:
        return np.mean(data, axis=0)
    elif units is False and time is True:
        return np.mean(data, axis=1)
    elif units is True and time is True:
        return np.mean(data)


def compound_mean(data):
    """
    Compute the mean of the compound/sum signal.
    Data is first summed across units and averaged across time.


    Parameters
    ----------
    data : numpy.ndarray
        1st axis unit, 2nd axis time


    Returns
    -------
    float
        time-averaged compound/sum signal


    Examples
    --------
    >>> compound_mean(np.array([[1, 2, 3], [4, 5, 6]]))
    7.0

    """

    return np.mean(np.sum(data, axis=0))


def variance(data, units=False, time=False):
    """
    Compute the variance of data across time, units or both.


    Parameters
    ---------- 
    data : numpy.ndarray 
        1st axis unit, 2nd axis time.
    units : bool 
        Variance across units
    time : bool
        Average over time


    Returns
    ---------- 
    if units=False and time=False: 
        Exception
    if units=True: 
        1 dim numpy.ndarray; time series     
    if time=True:  
        1 dim numpy.ndarray; series of single unit variances across time
    if units=True and time=True: 
        float; mean of single unit variances across time


    Examples
    ----------
    >>> variance(np.array([[1, 2, 3],[4, 5, 6]]), units=True)
    array([ 2.25,  2.25,  2.25])
    
    >>> variance(np.array([[1, 2, 3], [4, 5, 6]]), time=True)
    array([ 0.66666667,  0.66666667])
    
    >>> variance(np.array([[1, 2, 3], [4, 5, 6]]), units=True, time=True)
    0.66666666666666663

    """

    assert(units is not False or time is not False)
    
    if units is True and time is False:
        return np.var(data, axis=0)
    elif units is False and time is True:
        return np.var(data, axis=1)
    elif units is True and time is True:
        return np.mean(np.var(data, axis=1))


def compound_variance(data):
    """
    Compute the variance of the compound/sum signal.
    Data is first summed across units, then the variance across time
    is calculated.
    
    
    Parameters
    ----------
    data : numpy.ndarray
        1st axis unit, 2nd axis time
  

    Returns
    -------  
    float
        variance across time of compound/sum signal


    Examples
    --------
    >>> compound_variance(np.array([[1, 2, 3], [4, 5, 6]]))
    2.6666666666666665

    """

    return np.var(np.sum(data, axis=0))


def powerspec(data, tbin, Df=None, units=False, pointProcess=False):
    """
    Calculate (smoothed) power spectra of all timeseries in data.
    If units=True, power spectra are averaged across units.
    Note that averaging is done on power spectra rather than data.

    If pointProcess is True, power spectra are normalized by the length T of the
    time series.

 
    Parameters
    ----------
    data : numpy.ndarray
        1st axis unit, 2nd axis time.
    tbin : float
        Binsize in ms.
    Df : float/None, 
        Window width of sliding rectangular filter (smoothing),
        None is no smoothing.
    units : bool
        Average power spectrum.
    pointProcess : bool
        If set to True, powerspectrum is normalized to signal length T.


    Returns
    -------
    freq : tuple
        numpy.ndarray of frequencies.
    POW : tuple
        if units=False: 
            2 dim numpy.ndarray; 1st axis unit, 2nd axis frequency
        if units=True:  
            1 dim numpy.ndarray; frequency series

    
    Examples
    --------    
    >>> powerspec(np.array([analog_sig1, analog_sig2]), tbin, Df=Df)
    Out[1]: (freq,POW)
    >>> POW.shape
    Out[2]: (2,len(analog_sig1))

    >>> powerspec(np.array([analog_sig1, analog_sig2]), tbin, Df=Df, units=True)
    Out[1]: (freq,POW)
    >>> POW.shape
    Out[2]: (len(analog_sig1),)

    """

    freq, DATA = calculate_fft(data, tbin)
    df = freq[1] - freq[0]
    T = tbin * len(freq)
    POW = np.abs(DATA) ** 2
    if Df is not None:
        POW = [movav(x, Df, df) for x in POW]
        cut = int(Df / df)
        freq = freq[cut:]
        POW = np.array([x[cut:] for x in POW])
        POW = np.abs(POW)
    assert(len(freq) == len(POW[0]))
    if units is True:
        POW = mean(POW, units=units)
        assert(len(freq) == len(POW))
    if pointProcess:
        POW *= 1. / T * 1e3  # Normalization, power independent of T
    return freq, POW


def compound_powerspec(data, tbin, Df=None, pointProcess=False):
    """
    Calculate the power spectrum of the compound/sum signal.
    data is first summed across units, then the power spectrum is calculated.

    If pointProcess=True, power spectra are normalized by the length T of
    the time series.

    
    Parameters
    ----------
    data : numpy.ndarray, 
        1st axis unit, 2nd axis time
    tbin : float, 
        binsize in ms
    Df : float/None, 
        window width of sliding rectangular filter (smoothing),
        None -> no smoothing
    pointProcess : bool, 
        if set to True, powerspectrum is normalized to signal length T
                 


    Returns
    -------
    freq : tuple
        numpy.ndarray of frequencies
    POW : tuple
        1 dim numpy.ndarray, frequency series


    Examples
    --------
    >>> compound_powerspec(np.array([analog_sig1, analog_sig2]), tbin, Df=Df)
    Out[1]: (freq,POW)
    >>> POW.shape
    Out[2]: (len(analog_sig1),)

    """

    return powerspec([np.sum(data, axis=0)], tbin, Df=Df, units=True,
        pointProcess=pointProcess)


def crossspec(data, tbin, Df=None, units=False, pointProcess=False):
    """
    Calculate (smoothed) cross spectra of data.
    If `units`=True, cross spectra are averaged across units.
    Note that averaging is done on cross spectra rather than data.

    Cross spectra are normalized by the length T of the time series -> no
    scaling with T.

    If pointProcess=True, power spectra are normalized by the length T of the
    time series.


    Parameters
    ----------
    data : numpy.ndarray, 
        1st axis unit, 2nd axis time
    tbin : float, 
        binsize in ms
    Df : float/None, 
        window width of sliding rectangular filter (smoothing),
        None -> no smoothing
    units : bool, 
        average cross spectrum
    pointProcess : bool, 
        if set to True, cross spectrum is normalized to signal length T

    
    Returns
    -------
    freq : tuple
        numpy.ndarray of frequencies
    CRO : tuple
        if `units`=True: 1 dim numpy.ndarray; frequency series
        if `units`=False:3 dim numpy.ndarray; 1st axis first unit,
            2nd axis second unit, 3rd axis frequency


    Examples
    --------    
    >>> crossspec(np.array([analog_sig1, analog_sig2]), tbin, Df=Df)
    Out[1]: (freq,CRO)
    >>> CRO.shape
    Out[2]: (2,2,len(analog_sig1))

    >>> crossspec(np.array([analog_sig1, analog_sig2]), tbin, Df=Df, units=True)
    Out[1]: (freq,CRO)
    >>> CRO.shape
    Out[2]: (len(analog_sig1),)

    """

    N = len(data)
    if units is True:
        # smoothing and normalization take place in powerspec
        # and compound_powerspec
        freq, POW = powerspec(data, tbin, Df=Df, units=True)
        freq_com, CPOW = compound_powerspec(data, tbin, Df=Df)
        assert(len(freq) == len(freq_com))
        assert(np.min(freq) == np.min(freq_com))
        assert(np.max(freq) == np.max(freq_com))
        CRO = 1. / (1. * N * (N - 1.)) * (CPOW - 1. * N * POW)
        assert(len(freq) == len(CRO))
    else:
        freq, DATA = calculate_fft(data, tbin)
        T = tbin * len(freq)
        df = freq[1] - freq[0]
        if Df is not None:
            cut = int(Df / df)
            freq = freq[cut:]
        CRO = np.zeros((N, N, len(freq)), dtype=complex)
        for i in range(N):
            for j in range(i + 1):
                tempij = DATA[i] * DATA[j].conj()
                if Df is not None:
                    tempij = movav(tempij, Df, df)[cut:]
                CRO[i, j] = tempij
                CRO[j, i] = CRO[i, j].conj()
        assert(len(freq) == len(CRO[0, 0]))
        if pointProcess:
            CRO *= 1. / T * 1e3  # normalization
    return freq, CRO


def compound_crossspec(a_data, tbin, Df=None, pointProcess=False):
    """
    Calculate cross spectra of compound signals.
    a_data is a list of datasets (a_data = [data1,data2,...]).
    For each dataset in a_data, the compound signal is calculated
    and the crossspectra between these compound signals is computed.
    
    If pointProcess=True, power spectra are normalized by the length T of the
    time series.


    Parameters
    ----------
    a_data : list of numpy.ndarrays
        Array: 1st axis unit, 2nd axis time.     
    tbin : float
        Binsize in ms.
    Df : float/None, 
        Window width of sliding rectangular filter (smoothing),
        None -> no smoothing.
    pointProcess : bool
        If set to True, crossspectrum is normalized to signal length `T`
                
    Returns
    -------
    freq : tuple
        numpy.ndarray of frequencies.
    CRO : tuple
        3 dim numpy.ndarray; 1st axis first compound signal, 2nd axis second
        compound signal, 3rd axis frequency.


    Examples
    --------
    >>> compound_crossspec([np.array([analog_sig1, analog_sig2]),
                            np.array([analog_sig3,analog_sig4])], tbin, Df=Df)
    Out[1]: (freq,CRO)
    >>> CRO.shape
    Out[2]: (2,2,len(analog_sig1))

    """

    a_mdata = []
    for data in a_data:
        a_mdata.append(np.sum(data, axis=0))  # calculate compound signals
    return crossspec(np.array(a_mdata), tbin, Df, units=False,
                     pointProcess=pointProcess)


def autocorrfunc(freq, power):
    """
    Calculate autocorrelation function(s) for given power spectrum/spectra.


    Parameters
    ----------
    freq : numpy.ndarray
        1 dimensional array of frequencies.
    power : numpy.ndarray
        2 dimensional power spectra, 1st axis units, 2nd axis frequencies.


    Returns
    -------
    time : tuple
        1 dim numpy.ndarray of times.
    autof : tuple
        2 dim numpy.ndarray; autocorrelation functions, 1st axis units,
        2nd axis times.

    """
    tbin = 1. / (2. * np.max(freq)) * 1e3  # tbin in ms
    time = np.arange(-len(freq) / 2. + 1, len(freq) / 2. + 1) * tbin
    # T = max(time)
    multidata = False
    if len(np.shape(power)) > 1:
        multidata = True
    if multidata:
        N = len(power)
        autof = np.zeros((N, len(freq)))
        for i in range(N):
            raw_autof = np.real(np.fft.ifft(power[i]))
            mid = int(len(raw_autof) / 2.)
            autof[i] = np.hstack([raw_autof[mid + 1:], raw_autof[:mid + 1]])
        assert(len(time) == len(autof[0]))
    else:
        raw_autof = np.real(np.fft.ifft(power))
        mid = int(len(raw_autof) / 2.)
        autof = np.hstack([raw_autof[mid + 1:], raw_autof[:mid + 1]])
        assert(len(time) == len(autof))
    # autof *= T*1e-3 # normalization is done in powerspec()
    return time, autof


def crosscorrfunc(freq, cross):
    """
    Calculate crosscorrelation function(s) for given cross spectra.


    Parameters
    ----------
    freq : numpy.ndarray
        1 dimensional array of frequencies.
    cross : numpy.ndarray 
        2 dimensional array of cross spectra, 1st axis units, 2nd axis units,
        3rd axis frequencies.


    Returns
    -------
    time : tuple
        1 dim numpy.ndarray of times.
    crossf : tuple
        3 dim numpy.ndarray, crosscorrelation functions,
        1st axis first unit, 2nd axis second unit, 3rd axis times.

    """

    tbin = 1. / (2. * np.max(freq)) * 1e3  # tbin in ms
    time = np.arange(-len(freq) / 2. + 1, len(freq) / 2. + 1) * tbin
    # T = max(time)
    multidata = False
    # check whether cross contains many cross spectra
    if len(np.shape(cross)) > 1:
        multidata = True
    if multidata:
        N = len(cross)
        crossf = np.zeros((N, N, len(freq)))
        for i in range(N):
            for j in range(N):
                raw_crossf = np.real(np.fft.ifft(cross[i, j]))
                mid = int(len(raw_crossf) / 2.)
                crossf[i, j] = np.hstack(
                    [raw_crossf[mid + 1:], raw_crossf[:mid + 1]])
        assert(len(time) == len(crossf[0, 0]))
    else:
        raw_crossf = np.real(np.fft.ifft(cross))
        mid = int(len(raw_crossf) / 2.)
        crossf = np.hstack([raw_crossf[mid + 1:], raw_crossf[:mid + 1]])
        assert(len(time) == len(crossf))
    # crossf *= T*1e-3 # normalization happens in cross spectrum
    return time, crossf


def corrcoef(time, crossf, integration_window=0.):
    """
    Calculate the correlation coefficient for given auto- and crosscorrelation
    functions. Standard settings yield the zero lag correlation coefficient.
    Setting integration_window > 0 yields the correlation coefficient of
    integrated auto- and crosscorrelation functions. The correlation coefficient
    between a zero signal with any other signal is defined as 0.


    Parameters
    ----------
    time : numpy.ndarray
        1 dim array of times corresponding to signal.
    crossf : numpy.ndarray 
        Crosscorrelation functions, 1st axis first unit, 2nd axis second unit,
        3rd axis times.
    integration_window: float
        Size of the integration window.


    Returns
    -------
    cc : numpy.ndarray
        2 dim array of correlation coefficient between two units.

    """

    N = len(crossf)
    cc = np.zeros(np.shape(crossf)[:-1])
    tbin = abs(time[1] - time[0])
    lim = int(integration_window / tbin)
    
    if len(time)%2 == 0:
        mid = len(time)/2-1
    else:
        mid = np.floor(len(time)/2.)
    
    for i in range(N):
        ai = np.sum(crossf[i, i][mid - lim:mid + lim + 1])
        offset_autoi = np.mean(crossf[i,i][:mid-1])
        for j in range(N):
            cij = np.sum(crossf[i, j][mid - lim:mid + lim + 1])
            offset_cross = np.mean(crossf[i,j][:mid-1])
            aj = np.sum(crossf[j, j][mid - lim:mid + lim + 1])
            offset_autoj = np.mean(crossf[j,j][:mid-1])
            if ai > 0. and aj > 0.:
                cc[i, j] = (cij-offset_cross) / np.sqrt((ai-offset_autoi) * \
                    (aj-offset_autoj))
            else:
                cc[i, j] = 0.
    
    return cc


def coherence(freq, power, cross):
    """
    Calculate frequency resolved coherence for given power- and crossspectra.


    Parameters
    ----------
    freq : numpy.ndarray
        Frequencies, 1 dim array.
    power : numpy.ndarray
        Power spectra, 1st axis units, 2nd axis frequencies.
    cross : numpy.ndarray, 
        Cross spectra, 1st axis units, 2nd axis units, 3rd axis frequencies.


    Returns
    -------
    freq: tuple
        1 dim numpy.ndarray of frequencies.
    coh: tuple
        ndim 3 numpy.ndarray of coherences, 1st axis units, 2nd axis units,
        3rd axis frequencies.

    """

    N = len(power)
    coh = np.zeros(np.shape(cross))
    
    for i in range(N):
        for j in range(N):
            coh[i, j] = cross[i, j] / np.sqrt(power[i] * power[j])
    
    assert(len(freq) == len(coh[0, 0]))
    
    return freq, coh


def cv(data, units=False):
    """
    Calculate coefficient of variation (cv) of data. Mean and standard deviation
    are computed across time.


    Parameters
    ----------
    data : numpy.ndarray
        1st axis unit, 2nd axis time.
    units : bool
        Average `cv`.


    Returns
    -------
    numpy.ndarray
        If units=False, series of unit `cv`s.
    float
        If units=True, mean `cv` across units.


    Examples
    --------
    >>> cv(np.array([[1, 2, 3, 4, 5, 6], [11, 2, 3, 3, 4, 5]]))
    array([ 0.48795004,  0.63887656])

    >>> cv(np.array([[1, 2, 3, 4, 5, 6], [11, 2, 3, 3, 4, 5]]), units=True)
    0.56341330073710316

    """

    mu = mean(data, time=True)
    var = variance(data, time=True)
    cv = np.sqrt(var) / mu
    
    if units is True:
        return np.mean(cv)
    else:
        return cv


def fano(data, units=False):
    """
    Calculate fano factor (FF) of data. Mean and variance are computed across
    time.


    Parameters
    ----------   
    data : numpy.ndarray 
        1st axis unit, 2nd axis time.
    units : bool
        Average `FF`.


    Returns
    -------
    numpy.ndarray
        If units=False, series of unit FFs.
    float
        If units=True, mean FF across units.


    Examples
    --------
    >>> fano(np.array([[1, 2, 3, 4, 5, 6], [11, 2, 3, 3, 4, 5]]))
    array([ 0.83333333,  1.9047619 ])

    >>> fano(np.array([[1, 2, 3, 4, 5, 6], [11, 2, 3, 3, 4, 5]]), units=True)
    1.3690476190476191

    """
    mu = mean(data, time=True)
    var = variance(data, time=True)
    ff = var / mu
    
    if units is True:
        return np.mean(ff)
    else:
        return ff


if __name__ == "__main__":
    import doctest
    doctest.testmod()
    

